#include <myx/filesystem/whereami.hpp>

namespace whereami {

namespace {

#ifdef __cplusplus
extern "C" {
#endif

#include <cstdlib>

#ifndef WAI_NOINLINE
#if defined( _MSC_VER )
#define WAI_NOINLINE __declspec( noinline )
#elif defined( __GNUC__ )
#define WAI_NOINLINE __attribute__( ( noinline ) )
#else
#error unsupported compiler
#endif
#endif

#if defined( _MSC_VER )
#define WAI_RETURN_ADDRESS() _ReturnAddress()
#elif defined( __GNUC__ )
#define WAI_RETURN_ADDRESS() __builtin_extract_return_addr( __builtin_return_address( 0 ) )
#else
#error unsupported compiler
#endif


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if defined( __linux__ )
#include <linux/limits.h>
#else
#include <limits.h>
#endif
#ifndef __STDC_FORMAT_MACROS
#define __STDC_FORMAT_MACROS
#endif
#include <inttypes.h>

#if !defined( WAI_PROC_SELF_EXE )
#define WAI_PROC_SELF_EXE "/proc/self/exe"
#endif


int getExecutablePath( char* out, int capacity, int* dirname_length )
{
	char buffer[PATH_MAX];
	char* resolved = nullptr;
	int length = -1;

	for (;; )
	{
		resolved = realpath( WAI_PROC_SELF_EXE, buffer );
		if ( !resolved )
		{
			break;
		}

		length = static_cast< int >( strlen( resolved ) );
		if ( length <= capacity )
		{
			memcpy( out, resolved, static_cast< size_t >( length ) );

			if ( dirname_length )
			{
				int i;

				for ( i = length - 1; i >= 0; --i )
				{
					if ( out[i] == '/' )
					{
						*dirname_length = i;
						break;
					}
				}
			}
		}

		break;
	}

	return( length );
} // getExecutablePath


#if !defined( WAI_PROC_SELF_MAPS_RETRY )
#define WAI_PROC_SELF_MAPS_RETRY 5
#endif

#if !defined( WAI_PROC_SELF_MAPS )
#if defined( __sun )
#define WAI_PROC_SELF_MAPS "/proc/self/map"
#else
#define WAI_PROC_SELF_MAPS "/proc/self/maps"
#endif
#endif

#if defined( __ANDROID__ ) || defined( ANDROID )
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#endif


WAI_NOINLINE
int getModulePath( char* out, int capacity, int* dirname_length )
{
	int length = -1;
	FILE* maps = nullptr;

	for ( int r = 0; r < WAI_PROC_SELF_MAPS_RETRY; ++r )
	{
		maps = fopen( WAI_PROC_SELF_MAPS, "r" );
		if ( !maps )
		{
			break;
		}

		for (;; )
		{
			char buffer[PATH_MAX < 1024 ? 1024 : PATH_MAX];
			uint64_t low, high;
			char perms[5];
			uint64_t offset;
			uint32_t major, minor;
			char path[PATH_MAX];
			uint32_t inode;

			if ( !fgets( buffer, sizeof( buffer ), maps ) )
			{
				break;
			}

			if ( sscanf( buffer, "%" PRIx64 "-%" PRIx64 " %s %" PRIx64 " %x:%x %u %s\n", &low, &high, perms, &offset, &major, &minor, &inode, path ) == 8 )
			{
				uint64_t addr = reinterpret_cast< uintptr_t >( WAI_RETURN_ADDRESS() );
				if ( ( low <= addr ) && ( addr <= high ) )
				{
					char* resolved = nullptr;

					resolved = realpath( path, buffer );
					if ( resolved == nullptr )
					{
						break;
					}

					length = static_cast< int >( strlen( resolved ) );
					if ( length <= capacity )
					{
						memcpy( out, resolved, static_cast< size_t >( length ) );

						if ( dirname_length )
						{
							int i;

							for ( i = length - 1; i >= 0; --i )
							{
								if ( out[i] == '/' )
								{
									*dirname_length = i;
									break;
								}
							}
						}
					}

					break;
				}
			}
		}

		fclose( maps );
		maps = nullptr;

		if ( length != -1 )
		{
			break;
		}
	}

	if ( maps != nullptr )
	{
		fclose( maps );
	}

	return( length );
} // getModulePath


#ifdef __cplusplus

} // namespace
#endif

} // namespace


whereami_string_t whereami_path_t::dirname() const
{
	return( m_path.substr( 0, static_cast< size_t >( m_dirnameLength ) ) );
}


whereami_string_t whereami_path_t::basename() const
{
	return( m_path.substr( static_cast< size_t >( m_dirnameLength + 1 ) ) );
}


#if defined( WHEREAMI_CXX11 )
whereami_path_t::operator whereami_string_t() &&
{
	return( std::move( m_path ) );
}

whereami_path_t::operator whereami_string_t() const &
{
	return( m_path );
}


#else
whereami_path_t::operator const whereami_string_t&() const
{
	return( m_path );
}


#endif // if defined( WHEREAMI_CXX11 )

#if defined( WHEREAMI_CXX11 )
whereami_path_t::whereami_path_t( whereami_string_t&& path, int dirname_length ) noexcept
	: m_path( std::move( path ) ),
	m_dirnameLength( dirname_length )
{
}


#else
whereami_path_t::whereami_path_t( whereami_string_t& path, int dirname_length )
	: m_path( path ),
	m_dirnameLength( dirname_length )
{
}


#endif

#if !defined( WHEREAMI_DISABLE_OSTREAM )
std::ostream& operator<<( std::ostream& os, const whereami_path_t& path )
{
	return( os << path.m_path );
}


#endif


whereami_path_t getExecutablePath()
{
	whereami_string_t path;
	int dirname_length = -1;

	int length = getExecutablePath( nullptr, 0, nullptr );

	if ( length != -1 )
	{
		path.resize( static_cast< size_t >( length ) );
		getExecutablePath( &path[0], length, &dirname_length );
	}

	#if defined( WHEREAMI_CXX11 )
	return( whereami_path_t( std::move( path ), dirname_length ) );

	#else
	return( whereami_path_t( path, dirname_length ) );

	#endif
}


whereami_path_t getModulePath()
{
	whereami_string_t path;
	int dirname_length = -1;

	int length = getModulePath( nullptr, 0, nullptr );

	if ( length != -1 )
	{
		path.resize( static_cast< size_t >( length ) );
		getModulePath( &path[0], length, &dirname_length );
	}

	#if defined( WHEREAMI_CXX11 )
	return( whereami_path_t( std::move( path ), dirname_length ) );

	#else
	return( whereami_path_t( path, dirname_length ) );

	#endif
}

} // namespace whereami
